USE master
GO
DROP DATABASE ddr_db
GO
CREATE DATABASE ddr_db
GO
USE ddr_db
GO

------------------------------------------------------------------------------
-- The routing service and its data structures
--
------------------------------------------------------------------------------
CREATE TABLE [resolved_conversations] (
	inbound_conversation UNIQUEIDENTIFIER NOT NULL,
	outbound_conversation UNIQUEIDENTIFIER NOT NULL,
	PRIMARY KEY (inbound_conversation)
);

CREATE TABLE [routing_service_config] (
	service_name SYSNAME NOT NULL,
	classifier SYSNAME NOT NULL,
	PRIMARY KEY (service_name)
);
GO

CREATE QUEUE [routing_queue];
CREATE SERVICE [urn:rushi.desai.name/RoutingService]
	ON QUEUE [routing_queue] (
		[DEFAULT]
	);
GO

------------------------------------------------------------------------------
-- NAME: process_end_dialog_message
--
-- PURPOSE: If the inbound conversation is resolved, then end the
--	outbound conversation. In either case end the inbound conversation.
--
------------------------------------------------------------------------------
CREATE PROCEDURE process_end_dialog_message (
	@inbound_conversation UNIQUEIDENTIFIER)
AS
BEGIN
	DECLARE @outbound_conversation UNIQUEIDENTIFIER;
	SELECT @outbound_conversation = outbound_conversation
		FROM resolved_conversations
		WHERE inbound_conversation = @inbound_conversation;
	IF (@outbound_conversation IS NOT NULL)
	BEGIN
		END CONVERSATION @outbound_conversation;
	END
	END CONVERSATION @inbound_conversation;
	DELETE FROM resolved_conversations 
		WHERE inbound_conversation = @inbound_conversation
		OR outbound_conversation = @inbound_conversation;
END
GO

------------------------------------------------------------------------------
-- NAME: process_error_message
--
-- PURPOSE: If the inbound conversation is resolved, then end the outbound
--	conversation with an error wrapping the inbound error. Also end the 
--	inbound conversation.
--
------------------------------------------------------------------------------
CREATE PROCEDURE process_error_message (
	@inbound_conversation UNIQUEIDENTIFIER,
	@message_body VARBINARY(MAX))
AS
BEGIN
	DECLARE @outbound_conversation UNIQUEIDENTIFIER,
			@error_code INT,
			@error_description NVARCHAR(3000);
	SELECT @outbound_conversation = outbound_conversation
		FROM resolved_conversations
		WHERE inbound_conversation = @inbound_conversation;
	IF (@outbound_conversation IS NOT NULL)
	BEGIN
		SET @error_code = 10001; -- magic number indicating forwarded error
		SET @error_description = CONVERT(nvarchar(3000), @message_body);
		END CONVERSATION @outbound_conversation 
			WITH ERROR = @error_code DESCRIPTION = @error_description;
	END
	END CONVERSATION @inbound_conversation;
	DELETE FROM resolved_conversations 
		WHERE inbound_conversation = @inbound_conversation
		OR outbound_conversation = @inbound_conversation;
END
GO

------------------------------------------------------------------------------
-- NAME: process_message
--
-- PURPOSE: If the inbound conversation is resolved, then simply forward the
--	message. Else call the classifier procedure to resolve the conversation
--	and forward the message.
--
------------------------------------------------------------------------------
CREATE PROCEDURE process_message (
	@inbound_conversation UNIQUEIDENTIFIER,
	@contract_name NVARCHAR(256),
	@message_type_name NVARCHAR(256),
	@message_body VARBINARY(MAX))
AS
BEGIN
	DECLARE @outbound_conversation UNIQUEIDENTIFIER,
		@classifier SYSNAME,
		@to_service_name NVARCHAR(256);
	SELECT @outbound_conversation = outbound_conversation
		FROM resolved_conversations
		WHERE inbound_conversation = @inbound_conversation;
	IF (@outbound_conversation IS NULL)
	BEGIN;
		SELECT @classifier = classifier
			FROM routing_service_config
			WHERE service_name = 'urn:rushi.desai.name/RoutingService';
		EXEC @classifier @message_body, @to_service_name OUTPUT;
		IF (@to_service_name IS NULL)
		BEGIN
			END CONVERSATION @inbound_conversation WITH ERROR = 1 
				DESCRIPTION = N'Cannot resolve the message to target service.';
			RETURN;
		END
		BEGIN DIALOG @outbound_conversation
			FROM SERVICE [urn:rushi.desai.name/RoutingService]
			TO SERVICE @to_service_name
			ON CONTRACT @contract_name
			WITH ENCRYPTION = OFF;
		INSERT INTO resolved_conversations VALUES 
			(@inbound_conversation, @outbound_conversation);
		INSERT INTO resolved_conversations VALUES 
			(@outbound_conversation, @inbound_conversation);
	END;
	SEND ON CONVERSATION @outbound_conversation 
		MESSAGE TYPE @message_type_name (@message_body);
END
GO

------------------------------------------------------------------------------
-- NAME: route_message
--
-- PURPOSE: receive messages in a loop and route them as required.
--
------------------------------------------------------------------------------
CREATE PROCEDURE route_messages
AS
BEGIN
	DECLARE @message_type_name NVARCHAR(256),
			@contract_name NVARCHAR(256),
			@message_body VARBINARY(MAX),
			@inbound_conversation UNIQUEIDENTIFIER;
	WHILE (1=1)
	BEGIN 
		BEGIN TRANSACTION
		WAITFOR (RECEIVE TOP(1)
				@inbound_conversation = conversation_handle,
				@contract_name = service_contract_name,
				@message_type_name = message_type_name,
				@message_body = message_body
			FROM routing_queue), TIMEOUT 5000;
		
		IF (@@ROWCOUNT = 0)
		BEGIN
			ROLLBACK TRANSACTION;
			BREAK;
		END
		
		IF (@message_type_name = N'http://schemas.microsoft.com/SQL/ServiceBroker/Error')
		BEGIN
			EXEC process_error_message @inbound_conversation, @message_body;
		END
		ELSE IF (@message_type_name = N'http://schemas.microsoft.com/SQL/ServiceBroker/EndDialog')
		BEGIN
			EXEC process_end_dialog_message @inbound_conversation;
		END
		ELSE
		BEGIN
			EXEC process_message @inbound_conversation, @contract_name,
				@message_type_name, @message_body;
		END
		COMMIT
	END
END;
GO


------------------------------------------------------------------------------
-- Okay, let's turn on activation
--
------------------------------------------------------------------------------
ALTER QUEUE [routing_queue] WITH 
	ACTIVATION (
		STATUS = ON,
		PROCEDURE_NAME = route_messages,
		MAX_QUEUE_READERS = 4,
		EXECUTE AS SELF);

GO

------------------------------------------------------------------------------
-- Let's create some services
--
------------------------------------------------------------------------------
CREATE QUEUE [Aq];
CREATE SERVICE [A] ON QUEUE [Aq];
CREATE QUEUE [Bq];
CREATE SERVICE [B] ON QUEUE [Bq] ([DEFAULT]);
CREATE QUEUE [Cq];
CREATE SERVICE [C] ON QUEUE [Cq] ([DEFAULT]);
GO

------------------------------------------------------------------------------
-- Here's a very simple classifier that looks for service name inside an
-- XML message body
--
------------------------------------------------------------------------------

CREATE PROCEDURE sample_classifier (
	@message_body VARBINARY(MAX),
	@to_service_name NVARCHAR(256) OUTPUT)
AS
BEGIN
	DECLARE @xml_message XML;
	SET @xml_message = CONVERT(XML, @message_body);
	SET @to_service_name = @xml_message.value('(/message/toServiceName)[1]', 'nvarchar(256)');
END
GO

INSERT INTO [routing_service_config] VALUES (
	'urn:rushi.desai.name/RoutingService',
	'sample_classifier');
GO


------------------------------------------------------------------------------
-- Time to test it
--
------------------------------------------------------------------------------

-- Send two messages for C and two messages for B to the Routing Service
--
declare @dh1 uniqueidentifier, @dh2 uniqueidentifier;
begin tran
begin dialog @dh1
	from service A
	to service 'urn:rushi.desai.name/RoutingService'
	with encryption = off;
print 'Started conversation for C ''' + convert(nvarchar(80), @dh1) + '''';
send on conversation @dh1 (
N'<message>
	<toServiceName>C</toServiceName>
	<body>This is my first message to C.</body>
</message>');
send on conversation @dh1 (
N'<message>
	<body>This is my second message to C.</body>
</message>');
commit
begin tran
begin dialog @dh2
	from service A
	to service 'urn:rushi.desai.name/RoutingService'
	with encryption = off;
print 'Started conversation for B ''' + convert(nvarchar(80), @dh2) + '''';
send on conversation @dh2 (
N'<message>
	<toServiceName>B</toServiceName>
	<body>This is my first message to B.</body>
</message>');
send on conversation @dh2 (
N'<message>
	<body>This is my second message to B.</body>
</message>');
commit
go

-- Pause
waitfor delay '00:00:02';

-- View the resolved conversations
select * from [resolved_conversations]
go

-- Receive messages at B
declare @dh uniqueidentifier;
declare @body nvarchar(max);
while (1=1)
begin
	begin transaction
	waitfor (receive top(1) @dh=conversation_handle, @body=convert(nvarchar(max), message_body) 
		from Bq), timeout 5000;
	if (@@rowcount = 0)
	begin
		rollback;
		break;
	end
	print N'On conversation ''' + convert(nvarchar(80), @dh) + ''' received:';
	print @body;
	commit
end
if (@dh is not null)
begin
	end conversation @dh;
end
go


-- Receive messages at C
declare @dh uniqueidentifier;
declare @body nvarchar(max);
while (1=1)
begin
	begin transaction
	waitfor (receive top(1) @dh=conversation_handle, @body=convert(nvarchar(max), message_body) 
		from Cq), timeout 500;
	if (@@rowcount = 0)
	begin
		rollback;
		break;
	end
	print N'On conversation ''' + convert(nvarchar(80), @dh) + ''' received:';
	print @body;
	commit
end
if (@dh is not null)
begin
	end conversation @dh;
end
go

-- Receive messages at A
declare @dh uniqueidentifier;
declare @mt nvarchar(256);
while (1=1)
begin
	begin transaction
	waitfor (receive top(1) @dh=conversation_handle, @mt=message_type_name 
		from Aq), timeout 500;
	if (@@rowcount = 0)
	begin
		rollback;
		break;
	end
	if (@mt = 'http://schemas.microsoft.com/SQL/ServiceBroker/EndDialog')
	begin
		print N'Ended conversation ''' + convert(nvarchar(80), @dh) + '''';
		end conversation @dh;
	end
	commit
end
go
