USE master
GO
DROP DATABASE ddrr_db
GO
CREATE DATABASE ddrr_db
GO
USE ddrr_db
GO

------------------------------------------------------------------------------
-- The routing service and its metadata
--
------------------------------------------------------------------------------

CREATE TABLE [routing_service_config] (
	service_name SYSNAME NOT NULL,
	classifier SYSNAME NOT NULL,
	PRIMARY KEY (service_name)
);
GO

CREATE QUEUE [routing_queue];
CREATE SERVICE [urn:rushi.desai.name/RoutingService]
	ON QUEUE [routing_queue] (
		[DEFAULT]
	);
GO

CREATE MESSAGE TYPE [urn:rushi.desai.name/Redirect]
	VALIDATION = WELL_FORMED_XML;
GO

------------------------------------------------------------------------------
-- NAME: process_message
--
-- PURPOSE: If the inbound conversation is resolved, then simply forward the
--	message. Else call the classifier procedure to resolve the conversation
--	and forward the message.
--
------------------------------------------------------------------------------
CREATE PROCEDURE process_message (
	@inbound_conversation UNIQUEIDENTIFIER,
	@contract_name NVARCHAR(256),
	@message_type_name NVARCHAR(256),
	@message_body VARBINARY(MAX))
AS
BEGIN
	DECLARE @classifier SYSNAME,
		@to_service_name NVARCHAR(256),
		@to_broker_instance NVARCHAR(256),
		@reply_message NVARCHAR(1000);
	SELECT @classifier = classifier
		FROM routing_service_config
		WHERE service_name = 'urn:rushi.desai.name/RoutingService';
	EXEC @classifier @contract_name, @message_type_name, 
		@message_body, @to_service_name OUTPUT, @to_broker_instance OUTPUT;
	IF (@to_service_name IS NULL)
	BEGIN
		END CONVERSATION @inbound_conversation WITH ERROR = 1 
			DESCRIPTION = N'Cannot resolve the message to target service.';
		RETURN;
	END
	SET @reply_message = N'<redirectTo><serviceName>' + @to_service_name +
		N'</serviceName><brokerInstance>' + @to_broker_instance + 
		N'</brokerInstance></redirectTo>';
	SEND ON CONVERSATION @inbound_conversation MESSAGE TYPE
		[urn:rushi.desai.name/Redirect] (@reply_message);
	END CONVERSATION @inbound_conversation;
END
GO

------------------------------------------------------------------------------
-- NAME: route_message
--
-- PURPOSE: receive messages in a loop and route them as required.
--
------------------------------------------------------------------------------
CREATE PROCEDURE route_messages
AS
BEGIN
	DECLARE @message_type_name NVARCHAR(256),
			@contract_name NVARCHAR(256),
			@message_body VARBINARY(MAX),
			@inbound_conversation UNIQUEIDENTIFIER;
	WHILE (1=1)
	BEGIN 
		BEGIN TRANSACTION
		WAITFOR (RECEIVE TOP(1)
				@inbound_conversation = conversation_handle,
				@contract_name = service_contract_name,
				@message_type_name = message_type_name,
				@message_body = message_body
			FROM routing_queue), TIMEOUT 5000;
		
		IF (@@ROWCOUNT = 0)
		BEGIN
			ROLLBACK TRANSACTION;
			BREAK;
		END
		
		IF (@message_type_name = N'http://schemas.microsoft.com/SQL/ServiceBroker/Error'
				or
			@message_type_name = N'http://schemas.microsoft.com/SQL/ServiceBroker/EndDialog')
		BEGIN
			END CONVERSATION @inbound_conversation;
		END
		ELSE
		BEGIN
			EXEC process_message @inbound_conversation, @contract_name,
				@message_type_name, @message_body;
		END
		COMMIT
	END
END;
GO


------------------------------------------------------------------------------
-- Okay, let's turn on activation
--
------------------------------------------------------------------------------
ALTER QUEUE [routing_queue] WITH 
	ACTIVATION (
		STATUS = ON,
		PROCEDURE_NAME = route_messages,
		MAX_QUEUE_READERS = 4,
		EXECUTE AS SELF);

GO

------------------------------------------------------------------------------
-- Let's create some services
--
------------------------------------------------------------------------------

CREATE MESSAGE TYPE [TestMessage] VALIDATION = WELL_FORMED_XML;
GO

CREATE CONTRACT [TestContract] (
	[TestMessage] SENT BY ANY,
	[urn:rushi.desai.name/Redirect] SENT BY TARGET
);
GO

ALTER SERVICE [urn:rushi.desai.name/RoutingService] (
	ADD CONTRACT [TestContract])
GO

CREATE QUEUE [Aq];
CREATE SERVICE [A] ON QUEUE [Aq];
CREATE QUEUE [Bq];
CREATE SERVICE [B] ON QUEUE [Bq] ([TestContract]);
CREATE QUEUE [Cq];
CREATE SERVICE [C] ON QUEUE [Cq] ([TestContract]);
GO

------------------------------------------------------------------------------
-- Here's a very simple classifier that looks for service name inside an
-- XML message body
--
------------------------------------------------------------------------------

CREATE PROCEDURE sample_classifier (
	@contract_name NVARCHAR(256),
	@message_type_name NVARCHAR(256),
	@message_body VARBINARY(MAX),
	@to_service_name NVARCHAR(256) OUTPUT,
	@to_broker_instance NVARCHAR(256) OUTPUT)
AS
BEGIN
	DECLARE @xml_message XML;
	SET @xml_message = CONVERT(XML, @message_body);
	SET @to_service_name = @xml_message.value('(/message/toServiceName)[1]', 'nvarchar(256)');
	SET @to_broker_instance = 'CURRENT DATABASE';
END
GO

INSERT INTO [routing_service_config] VALUES (
	'urn:rushi.desai.name/RoutingService',
	'sample_classifier');
GO


------------------------------------------------------------------------------
-- Time to test it
--
------------------------------------------------------------------------------

-- Send two messages for C and two messages for B to the Routing Service
--
declare @dh1 uniqueidentifier, @dh2 uniqueidentifier,
	@replyMessage xml, @toService nvarchar(256),
	@toBrokerInstance nvarchar(256);
begin tran
begin dialog @dh1
	from service A
	to service 'urn:rushi.desai.name/RoutingService'
	on contract [TestContract]
	with encryption = off;
print 'Started conversation for C ''' + convert(nvarchar(80), @dh1) + '''';
send on conversation @dh1 message type [TestMessage] (
N'<message>
	<toServiceName>C</toServiceName>
	<body>This is my first message to C.</body>
</message>');
commit;
begin tran;
waitfor (
	receive top(1) @replyMessage=convert(xml, message_body) from Aq
	where conversation_handle = @dh1
), timeout 500;
if (@@rowcount > 0)
begin;
	end conversation @dh1;
	set @toService = @replyMessage.value('(/redirectTo/serviceName)[1]', 'nvarchar(256)');
	set @toBrokerInstance = @replyMessage.value('(/redirectTo/brokerInstance)[1]', 'nvarchar(256)');
	print 'Redirecting to ''' + @toService + ''', ''' + @toBrokerInstance + '''';
	begin dialog @dh1
		from service A
		to service @toService, @toBrokerInstance
		on contract [TestContract]
		with encryption = off;
	print 'Started conversation to C ''' + convert(nvarchar(80), @dh1) + '''';
	send on conversation @dh1 message type [TestMessage] (
N'<message>
	<toServiceName>C</toServiceName>
	<body>This is my first message to C.</body>
</message>');
end;
else
begin;
	print 'No reply from routing service!';
end;
commit;
go

-- Pause
waitfor delay '00:00:02';

-- Receive messages at C
declare @dh uniqueidentifier;
declare @body nvarchar(max);
while (1=1)
begin
	begin transaction
	waitfor (receive top(1) @dh=conversation_handle, @body=convert(nvarchar(max), message_body) 
		from Cq), timeout 500;
	if (@@rowcount = 0)
	begin
		rollback;
		break;
	end
	print N'On conversation ''' + convert(nvarchar(80), @dh) + ''' received:';
	print @body;
	commit
end
if (@dh is not null)
begin
	end conversation @dh;
end
go

-- Receive messages at A
declare @dh uniqueidentifier;
declare @mt nvarchar(256);
while (1=1)
begin
	begin transaction
	waitfor (receive top(1) @dh=conversation_handle, @mt=message_type_name 
		from Aq), timeout 500;
	if (@@rowcount = 0)
	begin
		rollback;
		break;
	end
	if (@mt = 'http://schemas.microsoft.com/SQL/ServiceBroker/EndDialog')
	begin
		print N'Ended conversation ''' + convert(nvarchar(80), @dh) + '''';
		end conversation @dh;
	end
	commit
end
go
